import logging
import pprint

from postgres.client.sql import SQLit

pp = pprint.PrettyPrinter(indent=2, width=140)

logging_obj = logging.getLogger(__name__)

def k8s_resources_tests_coverage(context, kind, label: str=None):
    """ Returns experiment coverage of k8s resources 
    (which experiments are available per namespace)
    """
    try:
        sql = SQLit()

        if label:
            report = sql.execute_as_dict(
            """SELECT * from k8s_resources_experiments_full_join_view
                WHERE kind = %s and name like %s and context = %s""",
                kind, label, context)
        else:
            report = sql.execute_as_dict(
            """SELECT * from k8s_resources_experiments_full_join_view
                WHERE kind = %s and context = %s""",
                kind, context)

        return report

    except Exception as e:
        logging_obj.error("Error getting report on resources: {}".format(e))
        return []

def get_last_test_status_for_resource(context, kind, label):
    """Returns the status of last tests (if any) for all resources (targets) in k8s of a
    given kind matching a given label in a given context"""

    try:
        sql = SQLit()
        # first we get all resources...
        resources = list()
        if label:
            resources = sql.execute_as_dict(
                """SELECT name, profile->'area' as area from k8s_resources
                    WHERE context = %s and kind = %s and name like %s""",
                    context, kind, label)
        else:
            resources = sql.execute_as_dict(
                """SELECT name, profile->'area' as area from k8s_resources
                    WHERE context = %s and kind = %s""",
                    context, kind)

        # now we get the report of last test performed on all resources 
        # this report does not count resources with no tests yet...
        report = sql.execute_as_dict("""
            SELECT DISTINCT k8s_resources.name,
                k8s_resources.profile -> 'area' AS area,
                sessions_targets.filepath,
                sessions_targets.title,
                sessions_targets.start_time,
                sessions_targets.status
            FROM k8s_resources
            FULL JOIN sessions_targets ON k8s_resources.name = sessions_targets.target)
            WHERE k8s_resources.context = %s AND
                sessions_targets.context = %s AND
                k8s_resources.kind = %s AND
                k8s_resources.name like %s AND
                sessions_targets.start_time IS NULL OR sessions_targets.start_time = ((
                SELECT max(es.start_time) AS max
                    FROM exp_session es
                    WHERE es.filepath = sessions_targets.filepath AND es.context = %s
                    GROUP BY es.filepath));
                """, context, context, kind, label, context)

        # remove deleted ones...
        for test in report:
            if test['status'] == 'deleted':
                report.remove(test)

        for resource in resources:
            if resource['name'] not in [ i['name'] for i in report ]:
                report.append({'name':resource['name'], 'area':resource['area']})

        return report
    except Exception as e: 
        logging_obj.error("Error getting report on resources: {}".format(e))
        return []

def get_test_history(context, name):
    """returns the history of all sessions regarding this experiment"""
    pass
