import logging
from postgres.client.sql import SQLit


class Resource:

    def __init__(self):
        self.sqlit = SQLit()
        self.logger = logging.getLogger(__name__)

    def get_tree_faster(self, **filter):
        """A much faster implementation, as it fetches all records at once.
        - roots := reference to the all tree roots found
        - children_dict := a dictionary with key the resource parent_id and
            value the resource parent dict
        """
        log = self.logger
        roots, children_dict = ([], {})
        where = dict([k, filter[k]] for k in ('layer', 'tag') if k in filter)
        try:
            resources = self.sqlit.select_as_dict(
                table='resources',
                sort='parent_id NULLS first, id',
                fetch=[
                    'id',
                    'parent_id',
                    'kind',
                    'name',
                    'creation_time',
                    'last_updated',
                    'layer',
                    'tag',
                    'status'
               ],
                **where
            )

            for res in resources:
                res['children'] = list()
                children_dict[res['id']] = res
                if res['parent_id'] != None and res['parent_id'] in children_dict:
                    children_dict[res['parent_id']]['children'].append(res)
                else:
                    roots.append(res)

            # okay, now we might end up with several roots...
            # we need to select actually the one that matches the passed filter...
            limited_roots = []
            for root in roots:
                limited_roots.append(dict([k, root[k]] for k in filter.keys()))

            for index, root in enumerate(limited_roots):
                if root == filter:
                    return roots[index]

            log.error("It seems we could not find a root resource matching filter...")
            return None

        except Exception as e:
            return None

    def get_tree(self, **filter):
        root = dict()
        root_rec = self.sqlit.select_as_dict(
            table='resources',
            fetch=[
                'id',
                'parent_id',
                'kind',
                'name',
                'creation_time',
                'last_updated',
                'layer',
                'tag',
                'status'
           ],
            **filter
        )
        if root_rec is not None:
            root = root_rec[0]
            self.logger.debug(f"Root: {root}")
            root['children'] = list()
            self.add_children(root)
        else:
            self.logger.error("No root element found")
        return root

    def add_children(self, parent):
        children = self.sqlit.select_as_dict(
            table='resources',
            fetch=[
                'id',
                'parent_id',
                'kind',
                'name',
                'creation_time',
                'last_updated',
                'layer',
                'tag',
                'status'
            ],
            **{'parent_id': parent['id']}
        )
        for child in children:
            child['children'] = list()
            self.add_children(child)
            parent['children'].append(child)
            self.logger.debug(f"Child {child} added to tree")

