import os
import logging
import json
from pprint import pformat
import http.client as http_client
from logging.handlers import RotatingFileHandler

from typing import Optional, List
from fastapi import FastAPI, Request, Query
from fastapi.middleware.cors import CORSMiddleware #https://fastapi.tiangolo.com/tutorial/cors/

from fastapi.responses import JSONResponse
from fastapi.responses import HTMLResponse

import p2_store.service
from p2_store.experiments import Experiment
from p2_store.sessions import Session
from p2_store.logs import Log
from p2_store.k8s_resources import K8sResource
from p2_store.campaigns import get_campaigns
from p2_store.resources import Resource
from postgres.client.sql import SQLit

logging.basicConfig(level=logging.DEBUG)
logging_obj = logging.getLogger(__name__)

handler = RotatingFileHandler(
    'server.log',
    maxBytes=1000000,
    backupCount=5
    )

# create formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# add formatter to ch
handler.setFormatter(formatter)

logging_obj.addHandler(handler)
logging_obj.info("Using p2_store version: {}".format(p2_store.service.__version__))

http_client.HTTPConnection.debuglevel = 1

# You must initialize logging, otherwise you'll not see debug output.
requests_log = logging.getLogger("requests.packages.urllib3")
requests_log.setLevel(logging.DEBUG)
requests_log.propagate = True

app = FastAPI()

# NOTE: we allow to provide content to ANY origin...
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_headers=["*"],
    allow_methods=["*"],
    expose_headers=["'Access-Control-Allow-Origin':'*'"]
)


@app.get('/')
def index():
    return '<br><br><h1> Welcome to Chaos. Interface for accessing and storing Chaos Experiments.</h1>'

@app.get('/store/version')
def getP2StoreVersion():
    return p2_store.service.__version__

@app.get("/url-list")
def get_all_urls():
    url_list = [{"path": route.path, "name": route.name} for route in app.routes]
    return url_list

@app.get("/store/campaigns")
def getCampaigns():
    try:
        records = get_campaigns( sort='id')
        logging_obj.info("getCampaigns: Records: {}".format(records))
        return records
    except Exception as err:
        logging_obj.error("GET session Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.get("/store/campaign/{id}")
def getCampaign(id):
    try:
        records = get_campaigns(id=int(id))
        if len(records) > 0:
            logging_obj.info("getCampaign: Campaign: {}".format(records[0]))
            return records[0]
        else:
            return JSONResponse(status_code=404)
    except Exception as err:
        logging_obj.error("GET session Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.get("/store/experiments")
def getExperiments(request: Request):
    try:
        filter = dict(request.query_params)
        exp_obj = Experiment()
        return exp_obj.getExperiments(**filter)
    except Exception as e:
        logging_obj.error("Error getting the list of experiments: {}".format(e))
        return []

@app.get("/store/experiments_as_dict")
def getExperimentsAsDict():
    try:
        exp_obj = Experiment()
        records = exp_obj.getExperimentsAsDict()
        return records
    except Exception as e:
        logging_obj.error("Error getting the list of experiments: {}".format(e))
        return {}

@app.get('/store/experiment/{filename}')
def getExperiment(filename):
    logging_obj.debug("Getting experiment: {}".format(filename))
    try:
        exp_obj = Experiment()
        return exp_obj.getExperiment(filename)
    except Exception as e:
        logging_obj.error("Error getting experiment: {}".format(filename))
        return JSONResponse(status_code=404, content={})

@app.delete('/store/experiment/{filename}')
def deleteEx(filename):
    logging_obj.debug("Deleting filename: {}".format(filename))
    try:
        exp_obj = Experiment()
        rowcount = exp_obj.deleteExperiment(filename)
        if (rowcount == 1):
            return filename
        else:
            return JSONResponse(status_code=404)
    except Exception as err:
        err_str = "ERROR: {}".format(err)
        logging_obj.error(err_str)
        return JSONResponse(status_code=500)

@app.post('/store/experiment/{filename}', status_code=201)
async def createExperiment(filename, request: Request):
    try:
        data = await request.json()
        logging_obj.debug("Request body: {}".format(data))
        exp_obj = Experiment()
        exp_obj.postExperiment(filename=filename, data=data)
        return {"filename":filename, "data":data}
    except Exception as err:
        logging_obj.error("ERROR: {}".format(err))
        return JSONResponse(status_code=500)

@app.put('/store/experiment/{filename}', status_code=200)
async def updateEx(filename, request: Request):
    try:
        data = await request.json()
        logging_obj.debug("[updateEx]:Request body: {}".format(data))
        exp_obj = Experiment()
        count = exp_obj.updateExperiment(filename=filename, data=data)
        if (count == 1):
            return exp_obj.getExperimentRec(filename)
        else:
            logging_obj.error("[updateEx]: experiment {} not found".format(filename))
            return JSONResponse(status_code=400)
    except Exception as err:
        logging_obj.error("ERROR: {}".format(err))
        return JSONResponse(status_code=500)

@app.post('/store/exp_session/{exp_id}', status_code=201)
async def create_session(exp_id, request: Request):
    try:
        data = await request.json()
        data['exp_id'] = exp_id
        logging_obj.debug("Creating session record: {}".format(data))
        session_obj = Session()
        session_obj.create_session(data)
        return data
    except Exception as err:
        logging_obj.error("POST Session Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.delete('/store/exp_session/{exp_id}')
async def delete_session(exp_id:str):
    """ Sets session status as deleted """
    try:
        sqlObj = SQLit()
        rows = sqlObj.update(
                table='exp_session',
                primary_key='exp_id',
                primary_key_value=exp_id,
                **{'status': 'deleted'}
        )
        if (rows == 1):
            logging_obj.info(f"Session {exp_id} status updated to deleted")
            return 1
        else:
            logging_obj.warning(f"Issues deleting session {exp_id} ...")
            return 0
    except Exception as err:
        logging_obj.error("DELETE Session {} Error: {}".format(exp_id, err))
        return JSONResponse(status_code=500)

@app.get('/store/exp_session')
def get_sessions(request: Request):
    try:
        filter = dict(request.query_params)
        logging_obj.info("[get_session]: filter: {}".format(filter))
        session = Session()
        records = session.get_sessions_as_dict( **filter )
        logging_obj.debug("Records: {}".format(records))
        return records
    except Exception as err:
        logging_obj.error("GET session Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.get('/store/list_sessions')
def list_sessions():
    try:
        session = Session()
        records = session.get_sessions_as_dict()
        logging_obj.debug("[list_sessions]: Records: {}".format(records))
        return records
    except Exception as err:
        logging_obj.error("[list_sessions]: Error: {}".format(err))
        return JSONResponse(status_code=500)


@app.put('/store/exp_session/{exp_id}')
async def update_session(exp_id, request: Request):
    try:
        data = await request.json()
        logging_obj.info("[update_session]: exp_id : {}".format(exp_id))
        logging_obj.info("[update_session]: payload: {}".format(data))

        session = Session()
        ret = session.update_session(exp_id, **data)
        if (ret == 1):
            return data
        else:
            return JSONResponse(status_code=400)
    except Exception as err:
        logging_obj.error("[update_session] Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.post('/store/exp_log/{exp_id}', status_code=201)
async def create_log(exp_id, request: Request):
    try:
        data = await request.json()
        logging_obj.debug("Creating for session {} log record: {}".format(exp_id, data))
        log_obj = Log()
        if (log_obj.create_log(exp_id, data) == 1):
            logging_obj.info("Log created")
            return data
        else:
            logging_obj.error("Error creating log")
            return JSONResponse(status_code=400)
    except Exception as err:
        logging_obj.error("POST Log Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.get('/store/exp_log/{exp_id}/{log_type}')
def get_log(exp_id: str, log_type: int):
    logging_obj.info("[get_log]: Getting log type {} for exp_id {}".format(log_type, exp_id))
    try:
        log_obj = Log()
        return log_obj.get_logs(exp_id=exp_id, log_type=log_type)
    except Exception as err:
        logging_obj.error("GET log Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.get('/store/k8s_resources')
def get_k8s_resources(request: Request, limit: int = None, offset: int = None):
    """ query params:
        - kind: kind of resource to get, mandatory
        - size: number of returned records, not mandatory, defaults to 1000
        - offset: index of next record to retrieve, not mandatory
    """
    try:
        filter = dict(request.query_params)
        resource = K8sResource()
        records = resource.get_resources(**filter, limit=limit, offset=offset)
        records = sorted(records, key=lambda record: record['name'])
        return records
    except Exception as err:
        logging_obj.error("GET k8s resources Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.post('/store/k8s_resource', status_code = 201)
async def add_resource(request: Request):
    try:
        data = await request.json()
        resource = K8sResource()
        count = resource.add_resource(data)
        if ( count == 1 ):
            return data
        else:
            logging_obj.error("[add_resource]:K8s Resource creation error")
            return JSONResponse(status_code=500)
    except Exception as err:
        logging_obj.error("POST K8s Resource Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.post('/store/cmd', status_code = 200)
async def execute_command(request: Request):
    try:
        data = await request.json()
        sql = SQLit()
        logging_obj.debug("execute_command: CMD: {}, attrs: {}".format(data['cmd'], data['attrs']))
        return sql.execute_as_dict(data['cmd'], *data['attrs'])
    except Exception as err:
        logging_obj.error("POST execute_command Error: {}".format(err))
        return JSONResponse(status_code=500)

@app.get('/store/k8s_resources_tests_status/{context}/{kind}')
async def k8s_resources_tests_status(context, kind):
    """ from k8s_resources we get the list of all resources
        for each resource:
            get the experiment name, title that address that resource
            get the last session start_time, status (ordered by start time desc) for that experiment in the context
            compile report with columns: resource name, exp title, last date, status
    """
    sql = SQLit()
    rows = sql.select('k8s_resources',
            context=context,
            kind=kind
            )
    # TODO


@app.post('/store/add_record', status_code = 201)
async def add_record(request: Request):
    """
    Adds a record. Payload:
    - table
    - record
    """
    try:
        data = await request.json()
        sql = SQLit()
        return sql.add(table=data['table'], **data['record'])
    except Exception as err:
        logging_obj.error("Data: ")
        logging_obj.error(pformat(data))
        logging_obj.error("Error adding a record: {}".format(err))
        return JSONResponse(status_code=500)


@app.post('/store/add_returning_key', status_code = 201)
async def add_returning_key(request: Request):
    """
    Adds a record and returns it's id. Payload:
    - table
    - key (serial key field)
    - record
    """
    try:
        data = await request.json()
        sql = SQLit()
        return sql.add_returning_id(
                table=data['table'],
                id=data['key'],
                **data['record'])
    except Exception as err:
        logging_obj.error("Data: ")
        logging_obj.error(pformat(data))
        logging_obj.error("Error adding a record: {}".format(err))
        return JSONResponse(status_code=500)


@app.get('/store/config/{component}')
def getConfig(component):
    logging_obj.debug("Getting configuration for component: {}".format(component))
    try:
        sql = SQLit()
        rows = sql.select('config', component=component)
        if len(rows) == 0:
           raise Exception("No configuration found for component: {}".format(component))
        return rows[0]['config']

    except Exception as e:
        logging_obj.error("""Error getting configuration for component {}:
            {}""".format(component, e))
        return JSONResponse(status_code=404, content={})

@app.get('/store/config/{component}/{tag}')
def getConfigWithTag(component, tag):
    logging_obj.debug(f"Getting configuration for component {component}, tag {tag}")
    try:
        sql = SQLit()
        rows = sql.select(
                'config', 
                component=component,
                tag=tag)
        if (len(rows) == 0):
           raise Exception(f"No configuration found for component {component}, tag {tag}")
        return rows[0]['config']

    except Exception as e:
        logging_obj.error(f"Error getting configuration for comp. {component}, tag {tag}: {e}")
        return JSONResponse(status_code=404, content={})


@app.post('/store/select/{table}', status_code = 200)
async def select(table: str, request: Request):
    """ Returns all records from the query specified. Params:
        table: (path) table to query from
        fetch: (multivalue query param): list of columns to fetch
            defaults to '*' if None
        limit: (q param) max number of rows, defaults to 10
        offset: (q param) index of first row retrieved
        sort: (q param) column [direction] to use to sort rows
    """
    data = dict()
    try:
        data = await request.json() or {}
    except:
        logging_obj.error("No data found")

    data['table'] = table
    logging_obj.debug(f"Params: {data}")
    try:
        sql = SQLit()
        records = sql.select_as_dict(**data)
        logging_obj.debug(f"Records: {records}")
        return records
    except Exception as e:
        logging_obj.error("Error select: {}".format(e))
        return JSONResponse(status_code=404, content={})


@app.get('/store/get_last_report/{type}/{tag}')
async def get_last_report(type, tag):
    """
    A route to get latest report
    given the type and tag
    """
    logging_obj.info(f"Getting latest report type {type} with tag {tag}")
    try:
        sql = SQLit()
        filter={
            'report_type': type,
            'tag': tag
            }
        report = sql.select_as_dict(
            table='reports',
            limit=1,
            offset=0,
            fetch=['id', 'title', 'description', 'params'],
            **filter
            )
        logging_obj.debug("Report: {}".format(report))
        if len(report) == 1:
            rows = sql.select_as_dict(
                table='report_entries',
                fetch=['body'],
                **{'report_id': report[0]['id']}
                ) 
            logging_obj.debug("Report rows: {}".format(rows))
            return {
                    'report': report,
                    'rows': rows
                    }
        else:
            logging_obj.error("Error getting report")
            return None
    except Exception as e:
        logging_obj.error("Error getting report: {}".format(e))
        return JSONResponse(status_code=404, content={})

@app.post('/store/log', status_code = 201)
async def add_log(request: Request):
    """ adds payload to logs table """
    try:
        data = await request.json()
        # logging_obj.info("[/log]: data:")
        # logging_obj.info(pformat(data))
        sql = SQLit()
        return sql.add(table='p2_logs', **data)
    except Exception as e:
        logging_obj.error(f"Adding log: {e}")
        return JSONResponse(status_code=500, content={})

@app.post('/store/update_on_conflict', status_code = 200)
async def update_on_conflict(request: Request):
    """ payload needs to be a dict with keys {
        table
        conflicting_cols
        record

        conflicting cols must be either a primary key or
        unique
    """
    try:
        data = await request.json()
        record = data['record']
        sql = SQLit()
        return sql.update_on_conflict(
                table=data['table'],
                conflicting_cols=data['conflicting_cols'],
                **record
                )
    except Exception as e:
        logging_obj.error("Error update_on_conflict: {}".format(e))
        return JSONResponse(status_code=500, content={})

@app.put('/store/update', status_code = 200)
async def update(request: Request):
    """ payload needs to be a dict with keys {
            table
            record
            where
        The where key should containt the primary key/value pairs.
        Note: it currently works with 1 primary key.
    """
    try:
        data = await request.json()
        record = data['record']
        where = data['where']
        sql = SQLit()
        return sql.update(
                table=data['table'],
                primary_key=list(where.keys())[0],
                primary_key_value=list(where.values())[0],
                **record
                )
    except Exception as e:
        logging_obj.error("Error updating: {}".format(e))
        return JSONResponse(status_code=500, content={})

@app.delete('/store/delete/{table}', status_code = 200)
async def delete(table: str, request: Request):
    """Deletes record pointed by key/value.
    payload:
        key
        value
    """
    try:
        payload = await request.json()
        sql = SQLit()
        return sql.delete(table=table, **payload)
    except Exception as e:
        logging_obj.error("Error deleting: {}".format(e))
        return JSONResponse(status_code=500, content={})

@app.get("/store/resources")
async def getResourcesTree(request: Request):
    try:
        params = dict(request.query_params) or {}
        logging_obj.debug("Resources params: {}".format(params))
        resource = Resource()
        return resource.get_tree(**params)
    except Exception as e:
        logging_obj.error(f"Error getting resources tree: {e}")
        return JSONResponse(status_code=500, content={})

@app.get("/store/resources/faster")
async def getResourcesTree(request: Request):
    try:
        params = dict(request.query_params) or {}
        logging_obj.debug("Resources params: {}".format(params))
        resource = Resource()
        return resource.get_tree_faster(**params)
    except Exception as e:
        logging_obj.error(f"Error getting resources tree: {e}")
        return JSONResponse(status_code=500, content={})
