import json
import logging
from croniter import croniter
from datetime import datetime

from postgres.client.sql import SQLit
from p2_store.experiments import Experiment

class Session:

    def __init__(self):
        self.sqlit = SQLit()
        self.log = logging.getLogger()

    def create_session( self, session ):
        for field in ('schedule', 'exp_id', 'filepath', 'campaign_id'):
            if field not in session: 
                raise Exception(f"{field} not in session record")

        if 'next_run_time' not in session:
            iter = croniter(session['schedule'], datetime.now())
            session['next_run_time'] = str(iter.get_next(datetime))

        if 'status' not in session:
            session['status'] = 'loaded'

        if 'profile' not in session or \
        ('profile' in session and 'target' not in session['profile']):
            session = self.update_profile(session)

        session['profile'] = json.dumps(session['profile'])

        if 'exp_id' in session:
            if self.sqlit.select_as_dict('exp_session', exp_id=session['exp_id']):
                exp_id = session.pop('exp_id')
                return self.sqlit.update('exp_session', 'exp_id', exp_id, **session)

        stmt = self.sqlit.insert_stmt('exp_session', session)
        return self.sqlit.execute(stmt, *list(session.values()))

    def update_profile(self, session: dict):
        exp_obj = Experiment()
        experiment = exp_obj.getExperimentRec(session['filepath'])
        if experiment is None: raise Exception("No experiment found")
        if 'profile' not in experiment or experiment['profile'] is None:
            self.log.warning("Experiment {} has no profile".format(experiment['filename']))
            session['profile'] = {}
            return session
        if 'profile' not in session:
            session['profile'] = dict()
        if 'target' in experiment['profile']:
            session['profile']['target'] = experiment['profile']['target']
        return session

    def get_sessions_paged( self, size=10, offset=0 ):
        """ 
        retrieves a max size of sessions starting from offset or start if none is passed
        if offset is 0 then retrieves the total number of rows as well
        """
        count = size

        if not offset:
            sql = "SELECT count(*) from exp_session;"
            recs = self.sqlit.execute(sql)
            count = recs[0][0]

        cmd = """
            SELECT * FROM exp_session 
                ORDER BY start_time DESC
                LIMIT %s
                OFFSET %s
            """
        return {
                "count": count,
                "records": self.sqlit.execute(cmd, size, offset)
        }


    def get_sessions(self, **filter):
        """given a filter as dict, retrieves all sessions that match it"""
        return self.sqlit.select('exp_session', **filter)

    def get_sessions_as_dict(self, **filter):
        """given a filter, retrieves all sessions that match it"""
        return self.sqlit.select_as_dict('exp_session', **filter)

    def update_session(self, exp_id, **attrs):
        return self.sqlit.update('exp_session', 'exp_id', exp_id, **attrs)

    def get_sessions_of_day(self, day: datetime):
        return self.get_sessions(
                dom="%02d" % day.day,
                month="%02d" % day.month,
                year=day.year
                )
